/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>

#include "gimageview.h"

#include "cursors.h"
#include "dnd.h"
#include "fileutil.h"
#include "gimv_image.h"
#include "gtk2-compat.h"
#include "icon.h"
#include "image_view.h"
#include "image_window.h"
#include "menu.h"
#include "nav-window.h"
#include "prefs.h"


typedef enum
{
   IMAGE_VIEW_NONE,
   IMAGE_VIEW_NEXT,
   IMAGE_VIEW_PREV,
   IMAGE_VIEW_POPUP,
   IMAGE_VIEW_ZOOM_IN,
   IMAGE_VIEW_ZOOM_OUT,
   IMAGE_VIEW_FIT,
   IMAGE_VIEW_ROTATE_CCW,
   IMAGE_VIEW_ROTATE_CW,
   IMAGE_VIEW_NAVWIN
} ImageViewPressType;


typedef enum {
   SET_LIST_SIGNAL,
   UNSET_LIST_SIGNAL,
   RENDERED_SIGNAL,
   TOGGLE_ASPECT_SIGNAL,
   TOGGLE_BUFFER_SIGNAL,
   LAST_SIGNAL
} ImageViewSignalType;


static void allocate_draw_buffer           (ImageView       *iv);
static void imageview_calc_image_size      (ImageView       *iv);
static void imageview_rotate_render        (ImageView       *iv,
					    guint            action);

/* call back functions for reference popup menu */
static void cb_keep_aspect                 (ImageView       *iv,
					    guint            action,
					    GtkWidget       *widget);
static void cb_zoom                        (ImageView       *iv,
					    ImgZoomType      zoom,
					    GtkWidget       *widget);
static void cb_rotate                      (ImageView       *iv,
					    guint            action,
					    GtkWidget       *widget);
static void cb_toggle_scrollbar            (ImageView       *iv,
					    guint            action,
					    GtkWidget       *widget);
static void cb_toggle_buffer               (ImageView       *iv,
					    guint            action,
					    GtkWidget       *widget);
static gint imageview_button_action        (ImageView       *iv,
					    GdkEventButton  *event,
					    gint             num);
static gint imageview_pressed              (ImageView       *iv,
					    GdkEventButton  *event,
					    gpointer         data);
static gint imageview_clicked              (ImageView       *iv,
					    GdkEventButton *event,
					    gpointer        data);

/* other call back functions */
static gint cb_image_configure             (GtkWidget         *widget,
					    GdkEventConfigure *event,
					    ImageView         *iv);
static gint cb_image_expose                (GtkWidget         *widget,
					    GdkEventExpose    *event,
					    ImageView         *iv);
static gint cb_image_key_press             (GtkWidget         *widget,
					    GdkEventKey       *event,
					    ImageView         *iv);
static gint cb_image_button_press          (GtkWidget         *widget,
					    GdkEventButton    *event,
					    ImageView         *iv);
static gint cb_image_button_release        (GtkWidget         *widget, 
					    GdkEventButton    *event,
					    ImageView         *iv);
static gint cb_image_motion_notify         (GtkWidget         *widget, 
					    GdkEventMotion    *event,
					    ImageView         *iv);

/* other private functions */
static void reset_scrollbar                (ImageView         *iv);


/* ImageView class func and variables */
static void imageview_class_init (ImageViewClass *klass);
static void imageview_init       (ImageView *iv);
#ifdef USE_GTK2
static void imageview_finalize   (GObject *object);
#else
static void imageview_finalize   (GtkObject *object);
#endif

static GtkWidgetClass *parent_class = NULL;
static gint imageview_signals[LAST_SIGNAL] = {0};


/* reference menu items */
GtkItemFactoryEntry imageview_popup_items [] =
{
   {N_("/tear"),                  NULL, NULL,                0, "<Tearoff>"},
   {N_("/_Zoom"),                 NULL, NULL,                0, "<Branch>"},
   {N_("/_Rotate"),               NULL, NULL,                0, "<Branch>"},
   {N_("/---"),                   NULL, NULL,                0, "<Separator>"},
   {N_("/Show _Scrollbar"),       NULL, cb_toggle_scrollbar, 0, "<ToggleItem>"},
   {N_("/---"),                   NULL, NULL,                0, "<Separator>"},
   {N_("/Memory _Buffer"),        NULL, cb_toggle_buffer,    0, "<ToggleItem>"},
   {NULL, NULL, NULL, 0, NULL},
};


/* for "Zoom" sub menu */
GtkItemFactoryEntry imageview_zoom_items [] =
{
   {N_("/tear"),                NULL,  NULL,            0,           "<Tearoff>"},
   {N_("/Zoom _In"),            NULL,  cb_zoom,         ZOOM_IN,     NULL},
   {N_("/Zoom _Out"),           NULL,  cb_zoom,         ZOOM_OUT,    NULL},
   {N_("/_Fit to Widnow"),      NULL,  cb_zoom,         ZOOM_FIT,    NULL},
   {N_("/Keep _aspect ratio"),  NULL,  cb_keep_aspect,  0,           "<ToggleItem>"},
   {N_("/---"),                 NULL,  NULL,            0,           "<Separator>"},
   {N_("/10%(_1)"),             NULL,  cb_zoom,         ZOOM_10,     NULL},
   {N_("/25%(_2)"),             NULL,  cb_zoom,         ZOOM_25,     NULL},
   {N_("/50%(_3)"),             NULL,  cb_zoom,         ZOOM_50,     NULL},
   {N_("/75%(_4)"),             NULL,  cb_zoom,         ZOOM_75,     NULL},
   {N_("/100%(_5)"),            NULL,  cb_zoom,         ZOOM_100,    NULL},
   {N_("/125%(_6)"),            NULL,  cb_zoom,         ZOOM_125,    NULL},
   {N_("/150%(_7)"),            NULL,  cb_zoom,         ZOOM_150,    NULL},
   {N_("/175%(_8)"),            NULL,  cb_zoom,         ZOOM_175,    NULL},
   {N_("/200%(_9)"),            NULL,  cb_zoom,         ZOOM_200,    NULL},
   {NULL, NULL, NULL, 0, NULL},
};


/* for "Rotate" sub menu */
GtkItemFactoryEntry imageview_rotate_items [] =
{
   {N_("/tear"),            NULL,  NULL,       0,           "<Tearoff>"},
   {N_("/Rotate 90 deg"),   NULL,  cb_rotate,  ROTATE_90,   NULL},
   {N_("/Rotate -90 deg"),  NULL,  cb_rotate,  ROTATE_270,  NULL},
   {N_("/Rotate 180 deg"),  NULL,  cb_rotate,  ROTATE_180,  NULL},
   {NULL, NULL, NULL, 0, NULL},
};


GList *ImageViewList;

static GdkPixmap *buffer = NULL;

static gboolean move_scrollbar_by_user = TRUE;


static void
allocate_draw_buffer (ImageView *iv)
{
   gint fwidth, fheight;

   if (!iv) return;

   if (buffer) gdk_pixmap_unref (buffer);
   imageview_get_image_frame_size (iv, &fwidth, &fheight);
   buffer = gdk_pixmap_new (iv->draw_area->window, fwidth, fheight, -1);
}



/*
 *  imageview_calc_image_size:
 *     @ calculate image size using specified scale and image frame size.
 *
 *  iv    : Pointer to ImageView struct.
 */
static void
imageview_calc_image_size (ImageView *iv)
{
   gint width, height;
   gfloat x_scale, y_scale;

   /* image scale */
   if (iv->rotate == 0 || iv->rotate == 2) {
      x_scale = iv->x_scale;
      y_scale = iv->y_scale;
   } else {
      x_scale = iv->y_scale;
      y_scale = iv->x_scale;
   }

   /* calculate image size */
   if (iv->fit_to_frame) {
      imageview_get_image_frame_size (iv, &width, &height);
      iv->x_scale = width  / (gfloat) gimv_image_width (iv->image) * 100;
      iv->y_scale = height / (gfloat) gimv_image_height (iv->image) * 100;

      if (iv->keep_aspect) {
	 if (iv->x_scale > iv->y_scale) {
	    iv->x_scale = iv->y_scale;
	    width = gimv_image_width (iv->image) * iv->x_scale / 100;
	 } else {
	    iv->y_scale = iv->x_scale;
	    height = gimv_image_height (iv->image) * iv->y_scale / 100;
	 }
      }

      iv->fit_to_frame = FALSE;
   } else {
      width  = gimv_image_width (iv->image)  * x_scale / 100;
      height = gimv_image_height (iv->image) * y_scale / 100;
   }

   /*
    * zoom out if image is too big & window is not scrollable in
    * fullscreen mode
    */
   /*
   win_width  = hadj->page_size;
   win_height = vadj->page_size;

   if (im->fullscreen && (width > win_width || height > win_height)) {
      gfloat width_ratio, height_ratio;

      width_ratio = (gdouble) width / hadj->page_size;
      height_ratio = (gdouble) height / vadj->page_size;
      if (width_ratio > height_ratio) {
	 width = win_width;
	 height = (gdouble) height / (gdouble) width_ratio;
      } else {
	 width = (gdouble) width / (gdouble) height_ratio;
	 height = win_height;
      }
   }
   */

   if (width > 8)
      iv->width = width;
   else
      iv->width = 8;
   if (height > 8)
      iv->height = height;
   else
      iv->height = 8;
}


/*
 *  imageview_rotate_render:
 *     @ Rotate image. Angle is relatively to old image.
 *
 *  iv    : Pointer to ImageView struct.
 *  angle : relative angle [1/90 deg]. Only 1, 2, 3 are recognized.
 */
static void
imageview_rotate_render (ImageView *iv, ImageAngle angle)
{
   switch (angle) {
   case ROTATE_90:
      iv->image = gimv_image_rotate_90 (iv->image, TRUE);
      break;
   case ROTATE_180:
      iv->image = gimv_image_rotate_180 (iv->image);
      break;
   case ROTATE_270:
      iv->image = gimv_image_rotate_90 (iv->image, FALSE);
      break;
   default:
      break;
   }
}


/******************************************************************************
 *
 *   Callback functions for reference popup menu.
 *
 ******************************************************************************/
/*
 *  cb_keep_aspect:
 *     @ Callback function for main menu item (/View/Zoom/Keep aspect ratio).
 *       Toggle keep aspect ratio or not when zoom image.
 *
 *  iv     : Pointer to ImageView struct.
 *  action :
 *  widget :
 */
static void
cb_keep_aspect (ImageView *iv, guint action, GtkWidget *widget)
{
   iv->keep_aspect = GTK_CHECK_MENU_ITEM(widget)->active;

   gtk_signal_emit (GTK_OBJECT(iv),
		    imageview_signals[TOGGLE_ASPECT_SIGNAL],
		    iv->keep_aspect);
}


/*
 *  cb_zoom:
 *     @ Callback function for main menu item (/View/Zoom/).
 *       Zoom image.
 *
 *  iv     : Pointer to ImageView struct.
 *  zoom   : Zoom type.
 *  widget :
 */
static void
cb_zoom (ImageView *iv, ImgZoomType zoom, GtkWidget *widget)
{
   imageview_zoom (iv, zoom, 0, 0);
}


/*
 *  cb_rotate:
 *     @ Callback function for main menu item (/View/Rotate/).
 *       Rotate image.
 *
 *  iv     : Pointer to ImageView struct.
 *  rotate : rotate angle (relativly) [1/90 deg]
 *  widget :
 */
static void
cb_rotate (ImageView *iv, ImageAngle rotate, GtkWidget *widget)
{
   guint angle;

   /* convert to absolute angle */
   angle = (iv->rotate + rotate) % 4;

   imageview_show_image (iv, angle);
}


static void
cb_toggle_scrollbar (ImageView *iv, guint action, GtkWidget *widget)
{
   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      imageview_show_scrollbar (iv);
   } else {
      imageview_hide_scrollbar (iv);
   }
}


/*
 *  cb_toggle_buffer:
 *     @ Callback function for popup menu item (/Memory Buffer).
 *       toggle Keep original image on memory or not.
 *
 *  iv     : Pointer to ImageView struct.
 *  action :
 *  widget :
 */
static void
cb_toggle_buffer (ImageView *iv, guint action, GtkWidget *widget)
{
   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      imageview_load_image_buf (iv);
      iv->buffer = TRUE;
   }
   else {
      iv->buffer = FALSE;
      imageview_free_image_buf (iv);
   }

   gtk_signal_emit (GTK_OBJECT(iv),
		    imageview_signals[TOGGLE_BUFFER_SIGNAL],
		    iv->buffer);
}


static gint
imageview_button_action (ImageView *iv, GdkEventButton *event, gint num)
{
   switch (abs (num)) {
   case IMAGE_VIEW_NEXT:
      imageview_next (iv);
      break;
   case IMAGE_VIEW_PREV:
      imageview_prev (iv);
      break;
   case IMAGE_VIEW_POPUP:
      imageview_popup_menu (iv, event);
      break;
   case IMAGE_VIEW_ZOOM_IN:
      imageview_zoom (iv, ZOOM_IN, 0, 0);
      break;
   case IMAGE_VIEW_ZOOM_OUT:
      imageview_zoom (iv, ZOOM_OUT, 0, 0);
      break;
   case IMAGE_VIEW_FIT:
      imageview_zoom (iv, ZOOM_FIT, 0, 0);
      break;
   case IMAGE_VIEW_ROTATE_CCW:
      imageview_rotate_ccw (iv);
      break;
   case IMAGE_VIEW_ROTATE_CW:
      imageview_rotate_cw (iv);
      break;
   case IMAGE_VIEW_NAVWIN:
      imageview_open_navwin (iv, event->x_root, event->y_root);
      break;
   default:
      break;
   }

   return TRUE;
}

static gint
imageview_pressed (ImageView *iv, GdkEventButton *event, gpointer data)
{
   gint num;

   g_return_val_if_fail (iv, TRUE);
   g_return_val_if_fail (event, TRUE);

   num = prefs_mouse_get_num_from_event (event, conf.imgview_mouse_button);
   if (num > 0)
      return imageview_button_action (iv, event, num);

   return TRUE;
}


static gint
imageview_clicked (ImageView *iv, GdkEventButton *event, gpointer data)
{
   gint num;

   g_return_val_if_fail (iv, TRUE);
   g_return_val_if_fail (event, TRUE);

   num = prefs_mouse_get_num_from_event (event, conf.imgview_mouse_button);
   if (num < 0)
      return imageview_button_action (iv, event, num);

   return TRUE;
}


/******************************************************************************
 *
 *   other callback functions.
 *
 ******************************************************************************/
/*
 *  cb_image_configure:
 *
 *  widget :
 *  event  :
 *  iv     :
 */
static gint
cb_image_configure (GtkWidget *widget, GdkEventConfigure *event, ImageView *iv)
{
   gint fwidth, fheight;

   imageview_get_image_frame_size (iv, &fwidth, &fheight);

   if (fwidth < iv->width) {
      if (iv->x_pos > 0 || iv->x_pos > fwidth || iv->x_pos < 0 - iv->width)
	 iv->x_pos = 0;
   } else {
      iv->x_pos = (fwidth - iv->width) / 2;
   }

   if (fheight < iv->height) {
      if (iv->y_pos > 0 || iv->y_pos > fheight || iv->y_pos < 0 - iv->height)
	 iv->y_pos = 0;
   } else {
      iv->y_pos = (fheight - iv->height) / 2;
   }

   imageview_draw_image (iv);
   reset_scrollbar (iv);

   return TRUE;
}


/*
 *  cb_image_expose:
 *     @ Callback function for Image (GtkDrawing Widget) expose event.
 *
 *  widget :
 *  event  :
 *  iv     : Pointer to ImageView struct.
 */
static gint
cb_image_expose (GtkWidget *widget, GdkEventExpose *event, ImageView *iv)
{
   if (iv->gdk_pixmap) {
      imageview_draw_image (iv);
   }
   return TRUE;
}


static gint
cb_image_key_press (GtkWidget *widget, GdkEventKey *event, ImageView *iv)
{
   guint keyval, popup_key;
   GdkModifierType modval, popup_mod;

   g_return_val_if_fail (iv, FALSE);

   keyval = event->keyval;
   modval = event->state;

   if (akey.common_popup_menu || *akey.common_popup_menu)
      gtk_accelerator_parse (akey.common_popup_menu, &popup_key, &popup_mod);
   else
      return FALSE;

   if (keyval == popup_key && (!popup_mod || (modval & popup_mod))) {
      imageview_popup_menu (iv, NULL);
   } else {
      switch (keyval) {
      case GDK_Left:
	 imageview_prev (iv);
	 return TRUE;
      case GDK_space:
      case GDK_Right:
	 imageview_next (iv);
	 return TRUE;
      case GDK_Up:
      case GDK_Down:
	 return TRUE;
      }
   }

   return FALSE;
}


/*
 *  cb_image_button_press:
 *     @ Callback function for mouse button press on Image (GtkDrawing Widget).
 *
 *  widget :
 *  event  :
 *  iv     : Pointer to ImageView struct.
 */
static gint
cb_image_button_press (GtkWidget *widget, GdkEventButton *event,
		       ImageView *iv)
{
   GdkCursor *cursor;
   gint retval;

   g_return_val_if_fail (iv, FALSE);

   iv->pressed = TRUE;
   iv->button  = event->button;

   gtk_widget_grab_focus (widget);

   if (iv->button_pressed_cb_func)
      iv->button_pressed_cb_func (iv, event, iv->button_pressed_cb_data);

   if (iv->dragging)
      return FALSE;

   if (event->button == 1) {   /* scroll image */
      if (!iv->gdk_pixmap)
	 return FALSE;

      cursor = cursor_get (widget->window, CURSOR_HAND_CLOSED);
      retval = gdk_pointer_grab (widget->window, FALSE,
				 (GDK_POINTER_MOTION_MASK
				  | GDK_POINTER_MOTION_HINT_MASK
				  | GDK_BUTTON_RELEASE_MASK),
				 NULL, cursor, event->time);
      gdk_cursor_destroy (cursor);

      if (retval != 0)
	 return FALSE;

      iv->drag_startx = event->x - iv->x_pos;
      iv->drag_starty = event->y - iv->y_pos;
      iv->x_pos_drag_start = iv->x_pos;
      iv->y_pos_drag_start = iv->y_pos;

      allocate_draw_buffer (iv);

      return TRUE;

   }

   return FALSE;
}


/*
 *  cb_image_button_release:
 *     @ Callback function for mouse button release on Image
 *       (GtkDrawing Widget).
 *
 *  widget :
 *  event  :
 *  iv     : Pointer to ImageView struct.
 */
static gint
cb_image_button_release  (GtkWidget *widget, GdkEventButton *event,
			  ImageView *iv)
{
   gint retval = TRUE;

   /*
   if (event->button != 1)
      return FALSE;
   */

   if (iv->button_released_cb_func)
      retval = iv->button_released_cb_func (iv, event, iv->button_released_cb_data);

   if(iv->pressed && !iv->dragging) {
      if (iv->button_clicked_cb_func)
	 retval = iv->button_clicked_cb_func (iv, event, iv->button_clicked_cb_data);
   }

   /*
   if (!im->dragging)
      gtk_signal_emit_by_name (GTK_OBJECT (im->event_box), "button_press_event");
   */

   if (buffer) {
      gdk_pixmap_unref (buffer);
      buffer = NULL;
   }

   iv->button   = 0;
   iv->pressed  = FALSE;
   iv->dragging = FALSE;

   gdk_pointer_ungrab (event->time);

   return retval;
}


/*
 *  cb_image_motion_notify:
 *     @ for image drag and scroll.
 *
 *  widget :
 *  event  :
 *  iv     : Pointer to ImageView struct.
 */
static gint
cb_image_motion_notify (GtkWidget *widget, GdkEventMotion *event,
			ImageView *iv)
{
   GdkModifierType mods;
   gint x, y, x_pos, y_pos, dx, dy;
   gint fwidth, fheight;

   if (!iv->pressed)
      return FALSE;

   if (event->is_hint)
      gdk_window_get_pointer (widget->window, &x, &y, &mods);

   x_pos = x - iv->drag_startx;
   y_pos = y - iv->drag_starty;
   dx = x_pos - iv->x_pos_drag_start;
   dy = y_pos - iv->y_pos_drag_start;

   if (!iv->dragging && (abs(dx) > 2 || abs (dy) > 2))
      iv->dragging = TRUE;

   /* scroll image */
   if (iv->button == 1) {
      if (event->is_hint)
	 gdk_window_get_pointer (widget->window, &x, &y, &mods);

      iv->x_pos = x_pos;
      iv->y_pos = y_pos;

      if (!conf.imgview_scroll_nolimit) {
	 imageview_get_image_frame_size (iv, &fwidth, &fheight);

	 if (iv->width <= fwidth) {
	    if (iv->x_pos < 0)
	       iv->x_pos = 0;
	    if (iv->x_pos > fwidth - iv->width)
	       iv->x_pos = fwidth - iv->width;
	 } else if (iv->x_pos < fwidth - iv->width) {
	    iv->x_pos = fwidth - iv->width;
	 } else if (iv->x_pos > 0) {
	    iv->x_pos = 0;
	 }

	 if (iv->height <= fheight) {
	    if (iv->y_pos < 0)
	       iv->y_pos = 0;
	    if (iv->y_pos > fheight - iv->height)
	       iv->y_pos = fheight - iv->height;
	 } else if (iv->y_pos < fheight - iv->height) {
	    iv->y_pos = fheight - iv->height;
	 } else if (iv->y_pos > 0) {
	    iv->y_pos = 0;
	 }
      }
   }

   imageview_draw_image (iv);
   reset_scrollbar (iv);

   return TRUE;
}


static void
cb_imageview_closed (GtkWidget *widget, ImageView *iv)
{
   ImageViewList = g_list_remove (ImageViewList, iv);

   if (iv->list_owner)
      imageview_remove_list (iv, iv->list_owner);
}


static void
cb_drag_data_received (GtkWidget *widget,
		       GdkDragContext *context,
		       gint x, gint y,
		       GtkSelectionData *seldata,
		       guint info,
		       guint32 time,
		       gpointer data)
{
   ImageView *iv = data;
   OpenFiles *files;
   GList *filelist;
   gchar *tmpstr;

   g_return_if_fail (iv || widget);

   filelist = dnd_get_file_list (seldata->data);

   if (filelist) {
      imageview_change_image_file (iv, (gchar *) filelist->data);
      imageview_show_image (iv, iv->rotate);
      tmpstr = filelist->data;
      filelist = g_list_remove (filelist, tmpstr);
      g_free (tmpstr);

      if (filelist) {
	 files = files_loader_new ();
	 files->filelist = filelist;
	 open_image_files_in_image_view (files);
	 files->filelist = NULL;
	 files_loader_delete (files);
      }
   }

   g_list_foreach (filelist, (GFunc) g_free, NULL);
   g_list_free (filelist);
}


void
cb_scrollbar_value_changed (GtkAdjustment *adj, ImageView *iv)
{
   g_return_if_fail (iv);

   if (!move_scrollbar_by_user) return;

   if (iv->width > iv->draw_area->allocation.width)
      iv->x_pos = 0 - iv->hadj->value;
   if (iv->height > iv->draw_area->allocation.height)
      iv->y_pos = 0 - iv->vadj->value;

   imageview_draw_image (iv);
}


void
imageview_open_navwin (ImageView *iv, gfloat x_root, gfloat y_root)
{
   GimvImage *image;
   GdkPixmap *pixmap;
   GdkBitmap *mask;
   gint src_width, src_height, dest_width, dest_height;

   g_return_if_fail (iv);
   if (!iv->gdk_pixmap) return;

   /* get pixmap for navigator */
   gdk_window_get_size (iv->gdk_pixmap, &src_width, &src_height);
   image = gimv_image_create_from_drawable (iv->gdk_pixmap, 0, 0,
					    src_width, src_height);
   g_return_if_fail (image);

   if (src_width > src_height) {
      dest_width  = NAV_WIN_SIZE;
      dest_height = src_height * NAV_WIN_SIZE / src_width;
   } else {
      dest_height = NAV_WIN_SIZE;
      dest_width  = src_width * NAV_WIN_SIZE / src_height;
   }

   gimv_image_scale_get_pixmap (image, dest_width, dest_height,
				&pixmap, &mask);
   if (!pixmap) goto ERROR;

   /* open navigate window */
   navwin_create (iv, x_root, y_root, pixmap, mask);

   /* free */
   gdk_pixmap_unref (pixmap);

 ERROR:
   gimv_image_kill (image);
}


static gint
cb_nav_button (GtkWidget *widget, GdkEventButton *event, ImageView *iv)
{
   g_return_val_if_fail (iv, FALSE);

   imageview_open_navwin (iv, event->x_root, event->y_root);

   return FALSE;
}


/******************************************************************************
 *
 *   Private functions.
 *
 ******************************************************************************/
static void
reset_scrollbar (ImageView *iv)
{
   g_return_if_fail (iv);
   g_return_if_fail (iv->draw_area);
   g_return_if_fail (iv->hadj);
   g_return_if_fail (iv->vadj);

   /* horizontal */
   if (iv->x_pos < 0)
      iv->hadj->value = 0 - iv->x_pos;
   else
      iv->hadj->value = 0;

   if (iv->width > iv->draw_area->allocation.width)
      iv->hadj->upper = iv->width;
   else
      iv->hadj->upper = iv->draw_area->allocation.width;
   iv->hadj->page_size = iv->draw_area->allocation.width;

   /* vertical */
   if (iv->y_pos < 0)
      iv->vadj->value = 0 - iv->y_pos;
   else
      iv->vadj->value = 0;

   if (iv->height > iv->draw_area->allocation.height)
      iv->vadj->upper = iv->height;
   else
      iv->vadj->upper = iv->draw_area->allocation.height;
   iv->vadj->page_size = iv->draw_area->allocation.height;

   move_scrollbar_by_user = FALSE;

   gtk_signal_emit_by_name (GTK_OBJECT(iv->hadj), "changed");
   gtk_signal_emit_by_name (GTK_OBJECT(iv->vadj), "changed");

   move_scrollbar_by_user = TRUE;
}


/******************************************************************************
 *
 *   Public functions.
 *
 ******************************************************************************/
/*
 *  imageview_create_zoom_menu:
 *     @ create referece "Zoom" sub menu.
 *
 *  window : Parent window.
 *  iv     :
 *  Return : GtkMenu widget.
 */
GtkWidget *
imageview_create_zoom_menu (GtkWidget *window, ImageView *iv)
{
   GtkWidget *menu;
   guint n_menu_items;

   n_menu_items = sizeof(imageview_zoom_items)
                     / sizeof(imageview_zoom_items[0]) - 1;
   menu = menu_create_items(window, imageview_zoom_items,
			    n_menu_items, "<ZoomSubMenu>", iv);
   menu_check_item_set_active (menu, "/Keep aspect ratio", iv->keep_aspect);

   return menu;
}


/*
 *  imageview_create_rotate_menu:
 *     @ create reference "Rotate" sub menu.
 *
 *  window : Parent window.
 *  iv     :
 *  Return : GtkMenu widget.
 */
GtkWidget *
imageview_create_rotate_menu (GtkWidget *window, ImageView *iv)
{
   GtkWidget *menu;
   guint n_menu_items;

   n_menu_items = sizeof(imageview_rotate_items)
                     / sizeof(imageview_rotate_items[0]) - 1;
   menu = menu_create_items(window, imageview_rotate_items,
			    n_menu_items, "<RotateSubMenu>", iv);

   return menu;
}


/*
 *  imageview_create_popup_menu:
 *     @ create reference popup menu.
 *
 *  window : Parent window.
 *  iv     :
 *  Return : GtkMenu widget.
 */
GtkWidget *
imageview_create_popup_menu (GtkWidget *window, ImageView *iv)
{
   guint n_menu_items;

   n_menu_items = sizeof(imageview_popup_items)
                     / sizeof(imageview_popup_items[0]) - 1;
   iv->imageview_popup = menu_create_items(window, imageview_popup_items,
					   n_menu_items, "<ViewSubMenu>", iv);

   iv->zoom_menu   = imageview_create_zoom_menu (window, iv);
   iv->rotate_menu = imageview_create_rotate_menu (window, iv);

   menu_set_submenu (iv->imageview_popup, "/Zoom",   iv->zoom_menu);
   menu_set_submenu (iv->imageview_popup, "/Rotate", iv->rotate_menu);

   menu_check_item_set_active (iv->imageview_popup, "/Show Scrollbar",
			       iv->show_scrollbar);
   menu_check_item_set_active (iv->imageview_popup, "/Memory Buffer",
			       iv->buffer);

   return iv->imageview_popup;
}


/*
 *  imageview_free_image_buf:
 *     @ free image buffer.
 *
 *  iv : Pointer to ImageView struct.
 */
void
imageview_free_image_buf (ImageView *iv)
{
   if (!iv)
      return;

   if (!iv->image) return;
   if (iv->buffer) return;

   gimv_image_kill (iv->image);
   iv->image = NULL;
}


/*
 *  imageview_load_image_buf:
 *     @ load image as GimvImage (GdkImlibImage, GdkPixbuf, ...)
 *       struct for rendering.
 *
 *  iv : Pointer to ImageView struct.
 */
void
imageview_load_image_buf (ImageView *iv)
{
   GimvImage *image, *rgb_image;
   gint bg_r = 255, bg_g = 255, bg_b = 255;
   const gchar *filename;

   g_return_if_fail (iv);

   if (!iv->info) return;
   if (!g_list_find (ImageViewList, iv)) return;

   if (iv->image) return;

   filename = image_info_get_path (iv->info);
   if (!filename || !*filename) return;

   /* load image buf */
   image = gimv_image_load_file (filename);
   if (!image) return;

   if (gimv_image_has_alpha (image)) {
      rgb_image = gimv_image_rgba2rgb (image, bg_r, bg_g, bg_b);
      if (rgb_image) {
	 gimv_image_kill (image);
	 image = rgb_image;
      }
   }

   if (!g_list_find (ImageViewList, iv)) {
      gimv_image_kill (image);
      return;
   } else {
      iv->image = image;
   }

   imageview_rotate_render (iv, iv->rotate);
}


/*
 *  imageview_change_image_file:
 *     @ Change to new image file.
 *
 *  iv      : Pointer to ImageView struct.
 *  newfile : new image file name.
 */
void
imageview_change_image_file (ImageView *iv, const gchar *newfile)
{
   g_return_if_fail (iv && newfile);

   if (!g_list_find (ImageViewList, iv)) return;

   /* free old image */
   if (iv->gdk_pixmap)
      gimv_image_free_pixmap_and_mask (iv->gdk_pixmap, iv->mask);
   iv->gdk_pixmap = NULL;
   iv->mask = NULL;

   if (iv->image)
      gimv_image_kill (iv->image);
   iv->image = NULL;

   /* free old image info */
   if (iv->info)
      image_info_unref (iv->info);

   /* reset fit_to_widnow */
   iv->fit_to_frame = iv->fit_to_frame_when_change_file;

   /* allocate memory for new image info*/
   iv->info = image_info_get (newfile);

   /* load image */
   imageview_load_image_buf (iv);

   if (!g_list_find (ImageViewList, iv)) {
      return;
   }

   if (!iv->image) {
      if (newfile && *newfile) {
	 if (!file_exists (newfile))
	    g_print (_("File not exist: %s\n"), newfile);
	 else
	    g_print(_("Not an image (or unsupported) file: %s\n"), newfile);
      }
      iv->image = NULL;

      /* clear draw area */
      imageview_draw_image (iv);
      reset_scrollbar (iv);

      return;
   }

   /* reset ImageInfo struct data */
   image_info_set_data_from_image (iv->info, iv->image);

   return;
}


/*
 *  imageview_zoom:
 *     @ set image scale, render, and display it.
 *
 *  iv      :
 *  zoom    : zoom type (ZOOM_IN, ZOOM_OUT, ZOOM_10, ZOOM_25, ...)
 *  x_scale : not used yet
 *  y_scale : not used yet
 */
void
imageview_zoom (ImageView *iv, ImgZoomType zoom, gint x_scale, gint y_scale)
{
   gint cx_pos, cy_pos, fwidth, fheight;
   gfloat src_x_scale, src_y_scale;

   g_return_if_fail (iv);

   src_x_scale = iv->x_scale;
   src_y_scale = iv->y_scale;

   switch (zoom) {
   case ZOOM_IN:
      if (iv->x_scale < IMG_MAX_SCALE && iv->y_scale < IMG_MAX_SCALE) {
	 iv->x_scale = iv->x_scale + IMG_MIN_SCALE;
	 iv->y_scale = iv->y_scale + IMG_MIN_SCALE;
      }
      break;
   case ZOOM_OUT:
      if (iv->x_scale > IMG_MIN_SCALE && iv->y_scale > IMG_MIN_SCALE) {
	 iv->x_scale = iv->x_scale - IMG_MIN_SCALE;
	 iv->y_scale = iv->y_scale - IMG_MIN_SCALE;
      }
      break;
   case ZOOM_FIT:
      iv->fit_to_frame = TRUE;
      break;
   case ZOOM_10:
      iv->x_scale = iv->y_scale =  10;
      break;
   case ZOOM_25:
      iv->x_scale = iv->y_scale =  25;
      break;
   case ZOOM_50:
      iv->x_scale = iv->y_scale =  50;
      break;
   case ZOOM_75:
      iv->x_scale = iv->y_scale =  75;
      break;
   case ZOOM_100:
      iv->x_scale = iv->y_scale = 100;
      break;
   case ZOOM_125:
      iv->x_scale = iv->y_scale = 125;
      break;
   case ZOOM_150:
      iv->x_scale = iv->y_scale = 150;
      break;
   case ZOOM_175:
      iv->x_scale = iv->y_scale = 175;
      break;
   case ZOOM_200:
      iv->x_scale = iv->y_scale = 200;
      break;	 
   case ZOOM_300:
      iv->x_scale = iv->y_scale = 300;
      break;	 
   default:
      break;
   }

   imageview_get_image_frame_size (iv, &fwidth, &fheight);

   cx_pos = (iv->x_pos - fwidth  / 2) * iv->x_scale / src_x_scale;
   cy_pos = (iv->y_pos - fheight / 2) * iv->y_scale / src_y_scale;

   iv->x_pos = cx_pos + fwidth  / 2;
   iv->y_pos = cy_pos + fheight / 2;

   if (zoom != ZOOM_FIT)
      iv->fit_to_frame = FALSE;

   imageview_show_image (iv, iv->rotate);
}


/*
 *  imageview_draw_image:
 *     @ draw image using double buffer.
 *
 *  iv : Pointer to ImageView struct.
 */
void
imageview_draw_image (ImageView *iv)
{
   GdkGC *bg_gc;
   gboolean free = FALSE;

   /* allocate buffer */
   if (!buffer) {
      allocate_draw_buffer (iv);
      free = TRUE;
   }

   /* fill background by default bg color */
   if (iv->bg_gc)
      bg_gc = iv->bg_gc;
   else
      bg_gc = iv->draw_area->style->bg_gc[GTK_WIDGET_STATE (iv->draw_area)];

   gdk_draw_rectangle (buffer, bg_gc, TRUE, 0, 0, -1, -1);

   /* draw image to buffer */
   if (iv->gdk_pixmap) {
      if (iv->mask) {
	 gdk_gc_set_clip_mask (iv->draw_area->style->black_gc, iv->mask);
	 gdk_gc_set_clip_origin (iv->draw_area->style->black_gc, iv->x_pos, iv->y_pos);
      }

      gdk_draw_pixmap (buffer,
		       iv->draw_area->style->black_gc,
		       iv->gdk_pixmap, 0, 0, iv->x_pos, iv->y_pos, -1, -1);

      if (iv->mask) {
	 gdk_gc_set_clip_mask (iv->draw_area->style->black_gc, NULL);
	 gdk_gc_set_clip_origin (iv->draw_area->style->black_gc, 0, 0);
      }
   }

   /* draw from buffer to foreground */
   gdk_draw_pixmap (iv->draw_area->window,
		    iv->draw_area->style->fg_gc[GTK_WIDGET_STATE (iv->draw_area)],
		    buffer, 0, 0, 0, 0, -1, -1);

   /* free buffer */
   if (free) {
      gdk_pixmap_unref (buffer);
      buffer = NULL;
   }
}


/*
 *  imageview_show_image:
 *     @ render image and display it.
 *
 *  iv    : Pointer to ImageView struct.
 *  abgle : Absolute image angle from original image.
 */
void
imageview_show_image (ImageView *iv, ImageAngle angle)
{
   ImageAngle rotate_angle;
   gint fwidth, fheight;

   imageview_load_image_buf (iv);

   if (!iv->image)
      goto func_end;

   /* rotate image */
   rotate_angle = (angle - iv->rotate) % 4;
   if (rotate_angle < 0)
      rotate_angle = rotate_angle + 4;

   iv->rotate = angle;
   imageview_rotate_render (iv, rotate_angle);

   /* set image size */
   imageview_calc_image_size (iv);

   /* image rendering */
   gimv_image_free_pixmap_and_mask (iv->gdk_pixmap, iv->mask);
   gimv_image_scale_get_pixmap (iv->image, iv->width, iv->height,
				&iv->gdk_pixmap, &iv->mask);

   /* reset image geometory */
   imageview_get_image_frame_size (iv, &fwidth, &fheight);
   if (fwidth < iv->width) {
      if (conf.imgview_scroll_nolimit) {
	 if (iv->x_pos > fwidth)
	    iv->x_pos = 0;
	 else if (iv->x_pos < 0 - iv->width)
	    iv->x_pos = 0 - iv->width + fwidth;
      } else {
	 if (iv->x_pos > 0)
	    iv->x_pos = 0;
	 else if (iv->x_pos < 0 - iv->width + fwidth)
	    iv->x_pos = 0 - iv->width + fwidth;
      }
   } else {
      iv->x_pos = (fwidth - iv->width) / 2;
   }

   if (fheight < iv->height) {
      if (conf.imgview_scroll_nolimit) {
	 if (iv->y_pos > fheight)
	    iv->y_pos = 0;
	 else if (iv->y_pos < 0 - iv->height)
	    iv->y_pos = 0 - iv->height + fheight;
      } else {
	 if (iv->y_pos > 0)
	    iv->y_pos = 0;
	 else if (iv->y_pos < 0 - iv->height + fheight)
	    iv->y_pos = 0 - iv->height + fheight;
      }
   } else {
      iv->y_pos = (fheight - iv->height) / 2;
   }

   /* draw image */
   imageview_draw_image (iv);
   reset_scrollbar (iv);

   imageview_free_image_buf (iv);

   /* set cursor */
   if (!iv->cursor)
      iv->cursor = cursor_get (iv->draw_area->window, CURSOR_HAND_OPEN);
   gdk_window_set_cursor (iv->draw_area->window, iv->cursor);

 func_end:
   gtk_signal_emit (GTK_OBJECT(iv), imageview_signals[RENDERED_SIGNAL]);
}


void
imageview_rotate_ccw (ImageView *iv)
{
   guint angle;

   g_return_if_fail (iv);

   /* convert to absolute angle */
   angle = (iv->rotate + 1) % 4;

   imageview_show_image (iv, angle);
}


void
imageview_rotate_cw (ImageView *iv)
{
   guint angle;

   g_return_if_fail (iv);

   /* convert to absolute angle */
   angle = (iv->rotate + 3) % 4;

   imageview_show_image (iv, angle);
}


void
imageview_moveto (ImageView *iv, gint x, gint y)
{
   g_return_if_fail (iv);

   iv->x_pos = 0 - x;
   iv->y_pos = 0 - y;

   imageview_draw_image (iv);
   reset_scrollbar (iv);
}


void
imageview_set_list (ImageView       *iv,
		    GList           *list,
		    GList           *current,
		    gpointer         list_owner,
		    ImageViewNextFn  next_fn,
		    gpointer         next_fn_data,
		    ImageViewPrevFn  prev_fn,
		    gpointer         prev_fn_data,
		    ImageViewNthFn   nth_fn,
		    gpointer         nth_fn_data,
		    ImageViewRemoveListFn remove_list_fn,
		    gpointer              remove_list_fn_data)
{
   g_return_if_fail (iv);
   g_return_if_fail (list);
   g_return_if_fail (current);

   if (iv->list_owner)
      imageview_remove_list (iv, iv->list_owner);

   iv->image_list = list;
   if (!current)
      iv->image_list = list;
   else
      iv->image_current = current;

   iv->list_owner   = list_owner;

   iv->next_fn      = next_fn;
   iv->next_fn_data = next_fn_data;

   iv->prev_fn      = prev_fn;
   iv->prev_fn_data = prev_fn_data;

   iv->nth_fn       = nth_fn;
   iv->nth_fn_data  = nth_fn_data;

   iv->remove_list_fn      = remove_list_fn;
   iv->remove_list_fn_data = remove_list_fn_data;

   gtk_signal_emit (GTK_OBJECT(iv), imageview_signals[SET_LIST_SIGNAL]);
}


void
imageview_remove_list (ImageView *iv, gpointer list_owner)
{
   g_return_if_fail (iv);

   if (iv->remove_list_fn && iv->list_owner == list_owner)
      iv->remove_list_fn (iv, iv->list_owner, iv->remove_list_fn_data);

   iv->image_list    = NULL;
   iv->image_current = NULL;
   iv->list_owner    = NULL;

   iv->next_fn       = NULL;
   iv->next_fn_data  = NULL;
   iv->prev_fn       = NULL;
   iv->prev_fn_data  = NULL;
   iv->nth_fn        = NULL;
   iv->nth_fn_data   = NULL;

   iv->remove_list_fn      = NULL;
   iv->remove_list_fn_data = NULL;

   gtk_signal_emit (GTK_OBJECT(iv), imageview_signals[UNSET_LIST_SIGNAL]);
}


void
imageview_popup_menu (ImageView *iv, GdkEventButton *event)
{
   guint button;
   guint32 time;
   GtkMenuPositionFunc pos_fn = NULL;

   g_return_if_fail (iv);
   /* g_return_if_fail (event); */

   if (event) {
      button = event->button;
      time = gdk_event_get_time ((GdkEvent *) event);
   } else {
      button = 0;
      time = GDK_CURRENT_TIME;
      pos_fn = menu_calc_popup_position;
   }

   if (iv->imageview_popup)
      gtk_menu_popup (GTK_MENU (iv->imageview_popup),
		      NULL, NULL,
		      pos_fn, iv->draw_area->window,
		      button, time);
}


void
imageview_next (ImageView *iv)
{
   g_return_if_fail (iv);
   if (!iv->next_fn) return;

   iv->image_current = iv->next_fn (iv, iv->list_owner,
				    iv->image_current, iv->next_fn_data);
}


void
imageview_prev (ImageView *iv)
{
   g_return_if_fail (iv);

   if (!iv->prev_fn) return;

   iv->image_current = iv->prev_fn (iv, iv->list_owner,
				    iv->image_current, iv->prev_fn_data);
}


void
imageview_nth (ImageView *iv, guint nth)
{
   g_return_if_fail (iv);

   if (!iv->nth_fn) return;

   iv->image_current = iv->nth_fn (iv, iv->list_owner,
				   iv->image_list, nth,
				   iv->nth_fn_data);
}


void
imageview_set_bg_color (ImageView *iv, GdkColor *color)
{
   g_return_if_fail (iv);
   g_return_if_fail (iv->draw_area);

   if (!iv->bg_gc) {
      iv->bg_gc = gdk_gc_new (iv->draw_area->window);
   }

   gdk_gc_set_foreground (iv->bg_gc, color);

   if (iv->image)
      imageview_draw_image (iv);
}


void
imageview_show_scrollbar (ImageView *iv)
{
   g_return_if_fail (iv);
   g_return_if_fail (iv->hscrollbar);
   g_return_if_fail (iv->vscrollbar);

   gtk_widget_show (iv->hscrollbar);
   gtk_widget_show (iv->vscrollbar);
   gtk_widget_show (iv->nav_button);

   iv->show_scrollbar = TRUE;
}


void
imageview_hide_scrollbar (ImageView *iv)
{
   g_return_if_fail (iv);
   g_return_if_fail (iv->hscrollbar);
   g_return_if_fail (iv->vscrollbar);

   gtk_widget_hide (iv->hscrollbar);
   gtk_widget_hide (iv->vscrollbar);
   gtk_widget_hide (iv->nav_button);

   iv->show_scrollbar = FALSE;
}


/*
 *  imageview_get_image_frame_size:
 *     @ Return frame (for drawing area) size.
 *
 *  iv     : Pointer to ImageView struct.
 *  width  : pointer to frame width  for return.
 *  height : pointer to frame height for return.
 */
void
imageview_get_image_frame_size (ImageView *iv, gint *width, gint *height)
{
   *width = *height = 0;

   *width  = iv->draw_area->allocation.width;
   *height = iv->draw_area->allocation.height;
}


void
imageview_memset (ImageView *iv)
{
   /* initialize struct data based on config */
   iv->image      = NULL;
   iv->gdk_pixmap = NULL;
   iv->mask       = NULL;

   iv->show_scrollbar  = FALSE;
   iv->hadj = iv->vadj = NULL;

   iv->fit_to_frame_when_change_file = conf.imgview_fit_image_to_win;
   iv->fit_to_frame   = conf.imgview_fit_image_to_win;
   iv->x_scale        = conf.imgview_scale;
   iv->y_scale        = conf.imgview_scale;
   iv->keep_aspect    = conf.imgview_keep_aspect;
   iv->rotate         = 0;
   iv->buffer         = conf.imgview_buffer;
   iv->show_scrollbar = conf.imgview_scrollbar;

   iv->bg_gc    = NULL;
   iv->bg_color = NULL;

   iv->cursor          = NULL;
   iv->imageview_popup = NULL;
   iv->zoom_menu       = NULL;
   iv->rotate_menu     = NULL;

   iv->x_pos = iv->y_pos = 0;

   iv->button_pressed_cb_func  = imageview_pressed;
   iv->button_pressed_cb_data  = NULL;
   iv->button_released_cb_func = NULL;
   iv->button_released_cb_data = NULL;
   iv->button_clicked_cb_func  = imageview_clicked;
   iv->button_clicked_cb_data  = NULL;

   iv->image_list    = NULL;
   iv->image_current = NULL;

   iv->next_fn = iv->next_fn_data = NULL;
   iv->prev_fn = iv->prev_fn_data = NULL;
   iv->nth_fn  = iv->nth_fn_data  = NULL;
}


GtkWidget*
imageview_new (const gchar *filename)
{
   ImageView *iv = gtk_type_new (imageview_get_type ());

   if (filename) {
      iv->info = image_info_get (filename);
   } else {
      iv->info = NULL;
   }

  /* load image */
   imageview_load_image_buf (iv);
   if (iv->image) {
      image_info_set_data_from_image (iv->info, iv->image);
   } else if (filename && *filename) {
      if (!file_exists (filename))
	 g_print (_("File not exist: %s\n"), filename);
      else if (conf.disp_filename_stdout)
	 g_print (_("Not an image (or unsupported) file: %s\n"), filename);
   }

   return GTK_WIDGET (iv);
}



/******************************************************************************
 *
 *   ImageView class funcs
 *
 ******************************************************************************/
guint
imageview_get_type ()
{
   static guint imageview_type = 0;

   if (!imageview_type) {
      static const GtkTypeInfo imageview_info = {
	 "ImageView",
	 sizeof (ImageView),
	 sizeof (ImageViewClass),
	 (GtkClassInitFunc) imageview_class_init,
	 (GtkObjectInitFunc) imageview_init,
	 NULL,
	 NULL,
	 (GtkClassInitFunc) NULL,
      };

      imageview_type = gtk_type_unique (gtk_vbox_get_type (), &imageview_info);
   }

   return imageview_type;
}


static void
imageview_class_init (ImageViewClass *klass)
{
   GtkObjectClass *object_class;

   object_class = (GtkObjectClass *) klass;
   parent_class = gtk_type_class (gtk_vbox_get_type ());

   imageview_signals[SET_LIST_SIGNAL]
      = gtk_signal_new ("set_list",
			GTK_RUN_FIRST,
			GTK_CLASS_TYPE(object_class),
			GTK_SIGNAL_OFFSET (ImageViewClass, set_list),
			gtk_signal_default_marshaller,
			GTK_TYPE_NONE, 0);

   imageview_signals[UNSET_LIST_SIGNAL]
      = gtk_signal_new ("unset_list",
			GTK_RUN_FIRST,
			GTK_CLASS_TYPE(object_class),
			GTK_SIGNAL_OFFSET (ImageViewClass, unset_list),
			gtk_signal_default_marshaller,
			GTK_TYPE_NONE, 0);

   imageview_signals[RENDERED_SIGNAL]
      = gtk_signal_new ("rendered",
			GTK_RUN_FIRST,
			GTK_CLASS_TYPE(object_class),
			GTK_SIGNAL_OFFSET (ImageViewClass, rendered),
			gtk_signal_default_marshaller,
			GTK_TYPE_NONE, 0);

   imageview_signals[TOGGLE_ASPECT_SIGNAL]
      = gtk_signal_new ("toggle_aspect",
			GTK_RUN_FIRST,
			GTK_CLASS_TYPE(object_class),
			GTK_SIGNAL_OFFSET (ImageViewClass, toggle_aspect),
			gtk_marshal_NONE__INT,
			GTK_TYPE_NONE, 1, GTK_TYPE_BOOL);

   imageview_signals[TOGGLE_BUFFER_SIGNAL]
      = gtk_signal_new ("toggle_buffer",
			GTK_RUN_FIRST,
			GTK_CLASS_TYPE(object_class),
			GTK_SIGNAL_OFFSET (ImageViewClass, toggle_buffer),
			gtk_marshal_NONE__INT,
			GTK_TYPE_NONE, 1, GTK_TYPE_BOOL);

   gtk_object_class_add_signals (object_class, imageview_signals, LAST_SIGNAL);

#ifdef USE_GTK2
   {
      GObjectClass *gobject_class      = G_OBJECT_CLASS (klass);
      gobject_class->finalize          = imageview_finalize;
   }
#else
  object_class->finalize               = imageview_finalize;
#endif

   klass->set_list      = NULL;
   klass->unset_list    = NULL;
   klass->rendered      = NULL;
   klass->toggle_aspect = NULL;
   klass->toggle_buffer = NULL;
}


static void
imageview_init (ImageView *iv)
{
   GtkWidget *event_box, *table;

   imageview_memset (iv);

   g_return_if_fail (iv);

   /* create widgets */
   table = gtk_table_new (2, 2, FALSE);
   gtk_widget_show (table);
   gtk_container_add (GTK_CONTAINER (iv), table);

   iv->draw_area = gtk_drawing_area_new ();
   gtk_widget_show (iv->draw_area);

   iv->hadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 10.0, 10.0, 0.0));
   iv->hscrollbar = gtk_hscrollbar_new (iv->hadj);
   gtk_widget_show (iv->hscrollbar);

   iv->vadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 10.0, 10.0, 0.0));
   iv->vscrollbar = gtk_vscrollbar_new (iv->vadj);
   gtk_widget_show (iv->vscrollbar);

   event_box = gtk_event_box_new ();
   gtk_widget_set_name (event_box, "/ImageView/NavWinButton");
   gtk_widget_show (event_box);
   iv->nav_button = icon_get_widget ("nav-button");
   gtk_container_add (GTK_CONTAINER (event_box), iv->nav_button);
   gtk_widget_show (iv->nav_button);

   gtk_table_attach (GTK_TABLE (table), iv->draw_area,
		     0, 1, 0, 1,
		     GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
   gtk_table_attach (GTK_TABLE (table), iv->vscrollbar,
		     1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
   gtk_table_attach (GTK_TABLE (table), iv->hscrollbar,
		     0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
   gtk_table_attach (GTK_TABLE (table), event_box,
		     1, 2, 1, 2, GTK_FILL, GTK_FILL, 0, 0);

   /* set signals */
   gtk_signal_connect (GTK_OBJECT (iv->draw_area), "configure_event",
		       GTK_SIGNAL_FUNC (cb_image_configure), iv);

   gtk_signal_connect (GTK_OBJECT (iv->draw_area), "expose_event",
		       GTK_SIGNAL_FUNC (cb_image_expose), iv);

   gtk_signal_connect_after (GTK_OBJECT (iv->draw_area), "key-press-event",
			     GTK_SIGNAL_FUNC(cb_image_key_press), iv);

   gtk_signal_connect (GTK_OBJECT (iv->draw_area), "button_press_event",
		       GTK_SIGNAL_FUNC (cb_image_button_press), iv);

   gtk_signal_connect (GTK_OBJECT (iv->draw_area), "button_release_event",
		       GTK_SIGNAL_FUNC (cb_image_button_release), iv);

   gtk_signal_connect (GTK_OBJECT (iv->draw_area), "motion_notify_event",
		       GTK_SIGNAL_FUNC (cb_image_motion_notify), iv);

   gtk_signal_connect (GTK_OBJECT (iv->draw_area), "destroy",
		       GTK_SIGNAL_FUNC (cb_imageview_closed), iv);

   gtk_signal_connect (GTK_OBJECT (iv->hadj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_scrollbar_value_changed), iv);

   gtk_signal_connect (GTK_OBJECT (iv->vadj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_scrollbar_value_changed), iv);

   gtk_signal_connect (GTK_OBJECT (event_box), "button_press_event",
		       GTK_SIGNAL_FUNC (cb_nav_button), iv);

   /* for droping file list */
   gtk_signal_connect(GTK_OBJECT (iv->draw_area), "drag_data_received",
		      GTK_SIGNAL_FUNC (cb_drag_data_received), iv);

   dnd_dest_set (iv->draw_area, dnd_types, dnd_types_num);

   /* set flags */
   GTK_WIDGET_SET_FLAGS (iv->draw_area, GTK_CAN_FOCUS);

   gtk_widget_add_events (iv->draw_area,
			  GDK_FOCUS_CHANGE
			  | GDK_BUTTON_PRESS_MASK | GDK_2BUTTON_PRESS
			  | GDK_KEY_PRESS | GDK_KEY_RELEASE
			  | GDK_BUTTON_RELEASE_MASK
			  | GDK_POINTER_MOTION_MASK
			  | GDK_POINTER_MOTION_HINT_MASK);

   /* add to list */
   ImageViewList = g_list_append (ImageViewList, iv);

   if (!iv->show_scrollbar) {
      imageview_hide_scrollbar (iv);
   }
}


static void
#ifdef USE_GTK2
imageview_finalize (GObject *object)
#else
imageview_finalize (GtkObject *object)
#endif
{
   ImageView *iv = IMAGEVIEW (object);

   gimv_image_free_pixmap_and_mask (iv->gdk_pixmap, iv->mask);
   iv->gdk_pixmap = NULL;
   iv->mask = NULL;

   if (iv->image)
      gimv_image_kill (iv->image);
   iv->image = NULL;

   if (iv->bg_gc)
      gdk_gc_destroy (iv->bg_gc);
   iv->bg_gc = NULL;

   if (iv->cursor)
      gdk_cursor_destroy (iv->cursor);

   /* decrease ref count of image info */
   if (iv->info)
      image_info_unref (iv->info);
   iv->info = NULL;

#ifdef USE_GTK2
   if (G_OBJECT_CLASS (parent_class)->finalize)
      G_OBJECT_CLASS (parent_class)->finalize (object);
#else
  if (GTK_OBJECT_CLASS (parent_class)->finalize)
     (*GTK_OBJECT_CLASS (parent_class)->finalize) (object);
#endif
}
