/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gdk/gdkkeysyms.h>

#include "gimageview.h"

#include "help.h"
#include "icon.h"
#include "image_view.h"
#include "image_window.h"
#include "menu.h"
#include "prefs.h"
#include "prefs_win.h"
#include "slideshow.h"

enum {
   IMG_FIRST,
   IMG_PREV,
   IMG_NEXT,
   IMG_LAST
};

GList *ImageWinList;
ImageWindow *shared_img_win;


static void  create_imageview_menus     (ImageWindow  *iw);
static GtkWidget *create_toolbar        (ImageWindow  *iw,
					 GtkWidget    *container);

static void imagewin_set_win_size       (ImageWindow  *iw);
static void imagewin_set_window_title   (ImageWindow  *iw);
static void imagewin_set_statusbar_info (ImageWindow  *iw);

/* callback functions for menubar */
static void cb_file_select          (gpointer     data,
				     guint        action,
				     GtkWidget   *widget);
static void cb_open_thumbnail       (gpointer     data,
				     guint        action,
				     GtkWidget   *widget);
static void cb_toggle_buffer        (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_window_close         (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_toggle_menubar       (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_toggle_player        (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_toggle_toolbar       (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_toggle_statusbar     (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_toggle_scrollbar     (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_toggle_fullscreen    (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_fit_to_image         (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);
static void cb_options              ();
static void cb_move_menu            (ImageWindow *iw,
				     guint        action,
				     GtkWidget   *widget);

/* callback functions for toolbar */
static void cb_toolbar_open_button  (GtkWidget *widget);
static void cb_toolbar_prefs_button (GtkWidget *widget);
static void cb_toolbar_prev_button  (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_next_button  (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_no_zoom      (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_zoom_in      (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_zoom_out     (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_zoom_fit     (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_zoom         (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_rotate       (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_keep_aspect  (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_fit_window   (GtkWidget *widget, ImageWindow *iw);
static void cb_toolbar_fullscreen   (GtkWidget *widget, ImageWindow *iw);
static gint cb_scale_spinner_key_press (GtkWidget *widget, 
					GdkEventKey *event,
					ImageWindow *iw);

/* other callback functions */
static void cb_window_closed (GtkWidget   *widget,
			      ImageWindow *iw);
static void cb_set_list      (ImageView   *iv,
			      ImageWindow *iw);
static void cb_unset_list    (ImageView   *iv,
			      ImageWindow *iw);
static void cb_rendered      (ImageView   *iv,
			      ImageWindow *iw);
static void cb_toggle_aspect (ImageView   *iv,
			      gboolean     keep_aspect,
			      ImageWindow *iw);

/* for main menu */
GtkItemFactoryEntry imagewin_menu_items[] =
{
   {N_("/_File"),                        NULL,         NULL,              0,  "<Branch>"},
   {N_("/_File/_Open..."),               NULL,         cb_file_select,    0,  NULL},
   {N_("/_File/Open _Thumbnail Window"), NULL,         cb_open_thumbnail, 0,  NULL},
   {N_("/_File/---"),                    NULL,         NULL,              0,  "<Separator>"},
   {N_("/_File/Memory _Buffer"),         NULL,         cb_toggle_buffer,  0,  "<ToggleItem>"},
   {N_("/_File/---"),                    NULL,         NULL,              0,  "<Separator>"},
   {N_("/_File/_Close"),                 NULL,         cb_window_close,   0,  NULL},
   {N_("/_File/_Quit"),                  NULL,         gimageview_quit,   0,  NULL},

   {N_("/_Edit"),                        NULL,         NULL,              0,  "<Branch>"},
   {N_("/_Edit/_Options..."),            NULL,         cb_options,        0,  NULL},

   {N_("/_View"),                        NULL,         NULL,              0,  "<Branch>"},

   {N_("/_Move"),                        NULL,         NULL,              0,  "<Branch>"},

   {N_("/_Help"),                        NULL,         NULL,              0,  "<LastBranch>"},
   {NULL, NULL, NULL, 0, NULL},
};


/* for "View" sub menu */
GtkItemFactoryEntry imagewin_view_items [] =
{
   {N_("/tear"),          NULL,  NULL,                 0,  "<Tearoff>"},
   {N_("/_Zoom"),         NULL,  NULL,                 0,  "<Branch>"},
   {N_("/_Rotate"),       NULL,  NULL,                 0,  "<Branch>"},
   {N_("/---"),           NULL,  NULL,                 0,  "<Separator>"},
   {N_("/_Menu Bar"),     NULL,  cb_toggle_menubar,    0,  "<ToggleItem>"},
   {N_("/_Tool Bar"),     NULL,  cb_toggle_toolbar,    0,  "<ToggleItem>"},
   {N_("/_Player"),       NULL,  cb_toggle_player,     0,  "<ToggleItem>"},
   {N_("/St_atus Bar"),   NULL,  cb_toggle_statusbar,  0,  "<ToggleItem>"},
   {N_("/_Scroll Bar"),   NULL,  cb_toggle_scrollbar,  0,  "<ToggleItem>"},
   {N_("/---"),           NULL,  NULL,                 0,  "<Separator>"},
   {N_("/F_ull Screen"),  NULL,  cb_toggle_fullscreen, 0,  "<ToggleItem>"},
   {N_("/_Fit to Image"), NULL,  cb_fit_to_image,      0,  NULL},
   {NULL, NULL, NULL, 0, NULL},
};


GtkItemFactoryEntry imagewin_move_items [] =
{
   {N_("/_Next"),        NULL,  cb_move_menu,  IMG_NEXT,  NULL},
   {N_("/_Previous"),    NULL,  cb_move_menu,  IMG_PREV,  NULL},
   {N_("/_First"),       NULL,  cb_move_menu,  IMG_FIRST, NULL},
   {N_("/_Last"),        NULL,  cb_move_menu,  IMG_LAST,  NULL},
   {NULL, NULL, NULL, 0, NULL},
};


/*
 *  create_imageview_menus:
 *     @ create main menu and popup menu.
 *
 *  im : Pointer to ImageWindow structure.
 */
static void
create_imageview_menus (ImageWindow *iw)
{
   guint n_menu_items;
   GtkWidget *help;

   /* main menu */
   n_menu_items = sizeof (imagewin_menu_items)
                     / sizeof (imagewin_menu_items[0]) - 1;
   iw->menubar = menubar_create (iw->window, imagewin_menu_items, n_menu_items,
				 "<ImageviewMain>", iw);
   gtk_container_add(GTK_CONTAINER(iw->menubar_handle), iw->menubar);
#ifndef USE_GTK2
   gtk_menu_bar_set_shadow_type (GTK_MENU_BAR(iw->menubar), GTK_SHADOW_NONE);
#endif
   gtk_widget_show (iw->menubar);

   /* sub menu */
   n_menu_items = sizeof(imagewin_view_items)
                     / sizeof(imagewin_view_items[0]) - 1;
   iw->view_menu = menu_create_items(iw->window, imagewin_view_items,
				     n_menu_items, "<ViewSubMenu>", iw);
   n_menu_items = sizeof(imagewin_move_items)
                     / sizeof(imagewin_move_items[0]) - 1;
   iw->move_menu = menu_create_items(iw->window, imagewin_move_items,
				     n_menu_items, "<MoveSubMenu>", iw);

   iw->zoom_menu   = imageview_create_zoom_menu (iw->window, iw->iv);
   iw->rotate_menu = imageview_create_rotate_menu (iw->window, iw->iv);
   help = gimvhelp_create_menu (iw->window);

   /* attach sub menus to parent menu */
   menu_set_submenu (iw->menubar,   "/View",   iw->view_menu);
   menu_set_submenu (iw->menubar,   "/Move",   iw->move_menu);
   menu_set_submenu (iw->menubar,   "/Help",   help);
   menu_set_submenu (iw->view_menu, "/Zoom",   iw->zoom_menu);
   menu_set_submenu (iw->view_menu, "/Rotate", iw->rotate_menu);

   /* popup menu */
   iw->iv->imageview_popup = iw->view_menu;

   /* initialize menubar check items */
   menu_check_item_set_active (iw->menubar,   "/File/Memory Buffer", iw->iv->buffer);

   menu_check_item_set_active (iw->view_menu, "/Menu Bar",    iw->show_menubar);
   menu_check_item_set_active (iw->view_menu, "/Tool Bar",    iw->show_toolbar);
   menu_check_item_set_active (iw->view_menu, "/Player",      iw->show_player);
   menu_check_item_set_active (iw->view_menu, "/Status Bar",  iw->show_statusbar);
   menu_check_item_set_active (iw->view_menu, "/Scroll Bar",  iw->iv->show_scrollbar);
   menu_check_item_set_active (iw->view_menu, "/Full Screen", iw->fullscreen);

   menu_item_set_sensitive (iw->move_menu, "/Next",     FALSE);
   menu_item_set_sensitive (iw->move_menu, "/Previous", FALSE);
   menu_item_set_sensitive (iw->move_menu, "/First",    FALSE);
   menu_item_set_sensitive (iw->move_menu, "/Last",     FALSE);
}


/*
 *  create_toolbar:
 *     @ create toolbar and buttons.
 *
 *  im        : Pointer to ImageWindow structure.
 *  container : Container to store toolbar.
 *  Return    : Pointer to toolbar widget.
 */
static GtkWidget *
create_toolbar (ImageWindow *iw, GtkWidget *container)
{
   GtkWidget *toolbar;
   GtkWidget *button;
   GtkWidget *iconw;
   GtkAdjustment *adj;
   GtkWidget *spinner;

#ifdef USE_GTK2
   toolbar = gtk_toolbar_new ();
#else
   toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_BOTH);
   gtk_toolbar_set_button_relief (GTK_TOOLBAR(toolbar), GTK_RELIEF_NONE);
   gtk_toolbar_set_space_style (GTK_TOOLBAR(toolbar), GTK_TOOLBAR_SPACE_LINE);
   gtk_toolbar_set_space_size  (GTK_TOOLBAR(toolbar), 16);
#endif


   /* file open button */
   iconw = icon_get_widget ("nfolder");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Open"),
				    _("File Open"),
				    _("File Open"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_open_button),
				    NULL);

   /* preference button */
   iconw = icon_get_widget ("prefs");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Prefs"),
				    _("Preference"),
				    _("Preference"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_prefs_button),
				    NULL);

   gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

   /* prev button */
   iconw = icon_get_widget ("back");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Prev"),
				    _("Previous Image"),
				    _("Previous Image"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_prev_button),
				    iw);
   iw->button.prev = button;
   /* gtk_widget_set_sensitive (button, FALSE); */

   /* next button */
   iconw = icon_get_widget ("forward");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Next"),
				    _("Next Image"),
				    _("Next Image"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_next_button),
				    iw);
   iw->button.next = button;
   /* gtk_widget_set_sensitive (button, FALSE); */

   gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

   /* no zoom button */
   iconw = icon_get_widget ("no_zoom");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("No Zoom"),
				    _("No Zoom"),
				    _("No Zoom"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_no_zoom),
				    iw);

   /* zoom in button */
   iconw = icon_get_widget ("zoom_in");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Zoom in"),
				    _("Zoom in"),
				    _("Zoom in"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_zoom_in),
				    iw);

   /* zoom out button */
   iconw = icon_get_widget ("zoom_out");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Zoom out"),
				    _("Zoom out"),
				    _("Zoom out"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_zoom_out),
				    iw);

   /* zoom fit button */
   iconw = icon_get_widget ("zoom_fit");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Zoom fit"),
				    _("Zoom fit"),
				    _("Zoom fit"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_zoom_fit),
				    iw);

   /* x scale spinner */
   adj = (GtkAdjustment *) gtk_adjustment_new (iw->iv->x_scale,
					       MIN_THUMB_SIZE, MAX_THUMB_SIZE,
					       1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_widget_set_name (spinner, "XScaleSpinner");
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
#ifndef USE_GTK2
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner),
				    GTK_SHADOW_NONE);
#endif
   gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar), spinner,
			      _("X Scale"), _("X Scale"));
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_toolbar_keep_aspect), iw);
   iw->button.xscale = spinner;
   gtk_signal_connect (GTK_OBJECT(spinner), "key-press-event",
		       GTK_SIGNAL_FUNC(cb_scale_spinner_key_press), iw);

   /* y scale spinner */
   adj = (GtkAdjustment *) gtk_adjustment_new (iw->iv->y_scale,
					       MIN_THUMB_SIZE, MAX_THUMB_SIZE,
					       1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_widget_set_name (spinner, "YScaleSpinner");
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
#ifndef USE_GTK2
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner),
				    GTK_SHADOW_NONE);
#endif
   gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar), spinner,
			      _("Y Scale"), _("Y Scale"));
   iw->button.yscale = spinner;

   /* zoom button */
   iconw = icon_get_widget ("zoom");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Zoom"),
				    _("Zoom"),
				    _("Zoom"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_zoom),
				    iw);

   gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

   /* rotate button */
   iconw = icon_get_widget ("rotate");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Rotate"),
				    _("Rotate CW"),
				    _("Rotate CW"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_rotate),
				    iw);

   gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

   /* resize button */
   iconw = icon_get_widget ("resize");
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Resize"),
				    _("Fit Window Size to Image"),
				    _("Fit Window Size to Image"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_fit_window),
				    iw);

   /* fullscreen button */
   iconw = icon_get_widget ("fullscreen");
   /* button = gtk_toggle_button_new (); */
   button = gtk_toolbar_append_item(GTK_TOOLBAR (toolbar),
				    _("Full"),
				    _("Fullscreen"),
				    _("Fullscreen"),
				    iconw,
				    GTK_SIGNAL_FUNC (cb_toolbar_fullscreen),
				    iw);

   gtk_widget_set_sensitive (iw->button.prev, FALSE);
   gtk_widget_set_sensitive (iw->button.next, FALSE);

   return toolbar;
}


/*
 *  create_player_toolbar:
 *     @ create toolbar and buttons.
 *
 *  im        : Pointer to ImageWindow structure.
 *  container : Container to store toolbar.
 *  Return    : Pointer to toolbar widget.
 */
static GtkWidget *
create_player_toolbar (ImageWindow *iw, GtkWidget *container)
{
   GtkWidget *toolbar;
   GtkWidget *button;
   GtkWidget *iconw;

#ifdef USE_GTK2
   toolbar = gtk_toolbar_new ();
#else
   toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_BOTH);
   gtk_toolbar_set_button_relief (GTK_TOOLBAR(toolbar), GTK_RELIEF_NONE);
   gtk_toolbar_set_space_style (GTK_TOOLBAR(toolbar), GTK_TOOLBAR_SPACE_LINE);
   gtk_toolbar_set_space_size  (GTK_TOOLBAR(toolbar), 16);
#endif


   /* previous button */
   iconw = icon_get_widget ("prev_t");
   button = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				     _("Prev"),
				     _("Previous"),
				     _("Previous"),
				     iconw,
				     NULL,
				     NULL);
   iw->player.prev = button;

   /* Reverse button */
   iconw = icon_get_widget ("rw");
   button = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				     _("RW"),
				     _("Reverse"),
				     _("Reverse"),
				     iconw,
				     NULL,
				     NULL);
   iw->player.rw = button;

   /* play button */
   iconw = icon_get_widget ("play");
   button = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				     _("Play"),
				     _("Play"),
				     _("Play"),
				     iconw,
				     NULL,
				     NULL);
   iw->player.play = button;

   /* pause button */
   iconw = icon_get_widget ("pause");
   button = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				     _("Pause"),
				     _("Pause"),
				     _("Pause"),
				     iconw,
				     NULL,
				     NULL);
   iw->player.pause = button;

   /* stop button */
   iconw = icon_get_widget ("stop2");
   button = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				     _("Stop"),
				     _("Stop"),
				     _("Stop"),
				     iconw,
				     NULL,
				     NULL);
   iw->player.stop = button;

   /* Forward button */
   iconw = icon_get_widget ("ff");
   button = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				     _("FF"),
				     _("Forward"),
				     _("Forward"),
				     iconw,
				     NULL,
				     NULL);
   iw->player.fw = button;

   /* Next button */
   iconw = icon_get_widget ("next_t");
   button = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				     _("Next"),
				     _("Next"),
				     _("Next"),
				     iconw,
				     NULL,
				     NULL);
   iw->player.next = button;

   gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

   /* Eject button */
   iconw = icon_get_widget ("eject");
   button = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				     _("Eject"),
				     _("Eject"),
				     _("Eject"),
				     iconw,
				     NULL,
				     NULL);
   iw->player.eject = button;

   gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

   return toolbar;
}


/*
 *  imageview_set_win_size:
 *     @ Set image window size to image size.
 *
 *  im    : Pointer to ImageWindow structure.
 */
static void
imagewin_set_win_size (ImageWindow *iw)
{
   gint x_size, y_size, width, height;
   gfloat x_scale, y_scale;

   if (!iw->window || iw->fullscreen)
      return;

   if (iw->iv->rotate == 0 || iw->iv->rotate == 2) {
      x_scale = iw->iv->x_scale;
      y_scale = iw->iv->y_scale;
      width =  iw->iv->info->width;
      height = iw->iv->info->height;
   } else {
      x_scale = iw->iv->y_scale;
      y_scale = iw->iv->x_scale;
      width  = iw->iv->info->height;
      height = iw->iv->info->width;
   }

   x_size = width  * x_scale / 100.0 + 0.5;
   y_size = height * y_scale / 100.0 + 0.5;

   if (iw->show_menubar)
      y_size += iw->menubar_handle->allocation.height;

   if (iw->show_toolbar)
      y_size += iw->toolbar_handle->allocation.height;

   if (iw->show_player)
      y_size += iw->player_handle->allocation.height;

   if (iw->show_statusbar)
      y_size += iw->status_bar_container->allocation.height;

   if (x_size < 1)
      x_size = conf.imgwin_width;
   if (y_size < 1)
      y_size = conf.imgwin_height;

   iw->iv->x_pos = 0;
   iw->iv->y_pos = 0;

   gdk_window_resize (iw->window->window, x_size, y_size);
   imageview_draw_image (iw->iv);
}


/*
 *  imageview_set_window_title:
 *     @ set window title (using image file name).
 *
 *  im : Pointer to ImageWindow structure.
 */
static void
imagewin_set_window_title (ImageWindow *iw)
{
   gchar buf[BUF_SIZE];
   const gchar *filename = NULL;
   gchar *dirname = NULL;

   g_return_if_fail (iw);
   if (!g_list_find (ImageWinList, iw)) return;

   g_return_if_fail (iw->iv);
   if (!g_list_find (ImageViewList, iw->iv)) return;

   if (iw->iv->info->filename) {
      filename = g_basename (iw->iv->info->filename);
      dirname = g_dirname (iw->iv->info->filename);
   }

   if (iw->iv->buffer)
      g_snprintf (buf, BUF_SIZE, _("%s (%s) - %s - "),
		  filename, dirname, PROG_NAME);
   else
      g_snprintf (buf, BUF_SIZE, _("* %s (%s) - %s - *"),
		  filename, dirname, PROG_NAME);

   gtk_window_set_title (GTK_WINDOW (iw->window), buf);

   g_free (dirname);
}


/*
 *  imageview_set_statusbar_info:
 *     @ set statubar strings.
 *
 *  im : Pointer to ImageWindow structure.
 */
static void
imagewin_set_statusbar_info (ImageWindow *iw)
{
   gchar buf[BUF_SIZE];

   g_return_if_fail (iw);
   if (!g_list_find (ImageWinList, iw)) return;

   g_return_if_fail (iw->iv);
   if (!g_list_find (ImageViewList, iw->iv)) return;

   if (iw->iv->buffer)
      g_snprintf (buf, BUF_SIZE, "%s (Buffer ON)", iw->iv->info->filename);
   else
      g_snprintf (buf, BUF_SIZE, "%s (Buffer OFF)", iw->iv->info->filename);

   gtk_statusbar_push (GTK_STATUSBAR (iw->status_bar1), 1, buf);
   g_snprintf (buf, BUF_SIZE, "%s %d x %d", iw->iv->info->format, 
	       iw->iv->info->width, iw->iv->info->height);
   gtk_statusbar_push (GTK_STATUSBAR (iw->status_bar2), 1, buf);
}



/******************************************************************************
 *
 *   Callback functions for menubar.
 *
 ******************************************************************************/
/*
 *  cb_file_select:
 *     @ Callback function for main menu item (/File/Open).
 *       Open file browser.
 *
 *  data   :
 *  action :
 *  widget :
 */
static void
cb_file_select (gpointer data, guint action, GtkWidget *widget)
{
   create_filebrowser (NULL);
}


/*
 *  cb_open_thumbnail:
 *     @ Callback function for main menu item (/File/Open Thumbnail Window).
 *       Open new thumbnail window.
 *
 *  data   :
 *  action :
 *  widget :
 */
static void
cb_open_thumbnail (gpointer data, guint action, GtkWidget *widget)
{
   thumbwin_open_window();
}


/*
 *  cb_toggle_buffer:
 *     @ Callback function for main menu item (/File/Memory Buffer).
 *       Keep original image on memory or not.
 *
 *  im     : Pointer to ImageWindow structure.
 *  action :
 *  widget :
 */
static void
cb_toggle_buffer (ImageWindow *iw, guint action, GtkWidget *widget)
{
   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      imageview_load_image_buf (iw->iv);
      iw->iv->buffer = TRUE;
   } else {
      iw->iv->buffer = FALSE;
      imageview_free_image_buf (iw->iv);
   }

   imagewin_set_window_title (iw);
   imagewin_set_statusbar_info (iw);
}


/*
 *  cb_window_close:
 *     @ Callback function for main menu item (/File/Close).
 *       Image window close.
 *
 *  im     : Pointer to ImageWindow structure.
 *  action :
 *  widget :
 */
static void
cb_window_close (ImageWindow *iw, guint action, GtkWidget *widget)
{
   if (!iw->window_creating)
      gtk_widget_destroy (iw->window);
}


/*
 *  cb_toggle_menubar:
 *     @ Callback function for main menu item (/View/Menu Bar).
 *       Show/Hide menu bar.
 *
 *  im     : Pointer to ImageWindow structure.
 *  action :
 *  widget :
 */
static void
cb_toggle_menubar (ImageWindow *iw, guint action, GtkWidget *widget)
{
   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      gtk_widget_show (iw->menubar_handle);
      iw->show_menubar = TRUE;
   } else {
      gtk_widget_hide (iw->menubar_handle);
      iw->show_menubar = FALSE;
   }

   if (!iw->window_creating)
      imageview_show_image (iw->iv, iw->iv->rotate);
}


/*
 *  cb_toggle_player:
 *     @ Callback function for main menu item (/View/Player).
 *       Show/Hide Player toolbar.
 *
 *  im     : Pointer to ImageWindow structure.
 *  action :
 *  widget :
 */
static void
cb_toggle_player (ImageWindow *iw, guint action, GtkWidget *widget)
{
   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      gtk_widget_show (iw->player_handle);
      iw->show_player = TRUE;
   } else {
      gtk_widget_hide (iw->player_handle);
      iw->show_player = FALSE;
   }

   if (!iw->window_creating)
      imageview_show_image (iw->iv, iw->iv->rotate);
}


/*
 *  cb_toggle_toolbar:
 *     @ Callback function for main menu item (/View/Tool Bar).
 *       Show/Hide tool bar.
 *
 *  im     : Pointer to ImageWindow structure.
 *  action :
 *  widget :
 */
static void
cb_toggle_toolbar (ImageWindow *iw, guint action, GtkWidget *widget)
{
   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      gtk_widget_show (iw->toolbar_handle);
      iw->show_toolbar = TRUE;
   } else {
      gtk_widget_hide (iw->toolbar_handle);
      iw->show_toolbar = FALSE;
   }

   if (!iw->window_creating)
      imageview_show_image (iw->iv, iw->iv->rotate);
}


static void
cb_toggle_statusbar (ImageWindow *iw, guint action, GtkWidget *widget)
{
   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      gtk_widget_show (iw->status_bar_container);
      iw->show_statusbar = TRUE;
   } else {
      gtk_widget_hide (iw->status_bar_container);
      iw->show_statusbar = FALSE;
   }

   if (!iw->window_creating)
      imageview_show_image (iw->iv, iw->iv->rotate);
}


/*
 *  cb_toggle_statusbar:
 *     @ Callback function for main menu item (/View/Status Bar).
 *       Show/Hide status bar.
 *
 *  im     : Pointer to ImageWindow structure.
 *  action :
 *  widget :
 */
static void
cb_toggle_scrollbar (ImageWindow *iw, guint action, GtkWidget *widget)
{
   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      imageview_show_scrollbar (iw->iv);
   } else {
      imageview_hide_scrollbar (iw->iv);
   }
}


/*
 *  cb_toggle_fullscreen:
 *     @ Callback function for main menu item (/View/Full Screen).
 *
 *  im     : Pointer to ImageWindow structure.
 *  action :
 *  widget :
 */
static void
cb_toggle_fullscreen (ImageWindow *iw, guint action, GtkWidget *widget)
{
   GdkWindow *gdk_window = iw->window->window;
   gint client_x, client_y, root_x, root_y;
   gint width, height;

   if (GTK_CHECK_MENU_ITEM(widget)->active) {
      gdk_window_get_origin (gdk_window, &root_x, &root_y);
      gdk_window_get_geometry (gdk_window, &client_x, &client_y,
			       &width, &height, NULL);

      gdk_window_move_resize (gdk_window, -client_x, -client_y,
			      gdk_screen_width (), gdk_screen_height ());
      gdk_window_raise (gdk_window);

      iw->fullscreen = TRUE;
      iw->win_x = root_x - client_x;
      iw->win_y = root_y - client_y;
      iw->win_width  = width;
      iw->win_height = height;
   } else {
      gdk_window_move_resize (gdk_window, iw->win_x, iw->win_y,
			      iw->win_width, iw->win_height);
      iw->fullscreen = FALSE;
   }
}


/*
 *  cb_fit_to_image:
 *     @ Callback function for main menu item (/View/Full Screen).
 *     @ Fit window size to image.
 */
static void
cb_fit_to_image (ImageWindow *iw, guint action, GtkWidget *widget) {
   imagewin_set_win_size (iw);
}


/*
 *  cb_options:
 *     @ Callback function for main menu item (/Preference/Options).
 *     @ Open preference window.
 */
static void
cb_options ()
{
   prefs_open_window ();
}


static void
cb_move_menu (ImageWindow *iw,
	      guint        action,
	      GtkWidget   *widget)
{
   gint num;

   g_return_if_fail (iw);

   switch (action) {
   case IMG_FIRST:
      imageview_nth (iw->iv, 0);
      break;
   case IMG_PREV:
      imageview_prev (iw->iv);
      break;
   case IMG_NEXT:
      imageview_next (iw->iv);
      break;
   case IMG_LAST:
      num = g_list_length (iw->iv->image_list);
      imageview_nth (iw->iv, num - 1);
      break;
   default:
      break;
   }
}


/******************************************************************************
 *
 *  callback functions for toolbar.
 *
 ******************************************************************************/
/*
 *  cb_toolbar_open_button:
 *     @ Callback function for open button.
 *     @ Open file browser. 
 *
 *  widget :
 *  im     : Pointer to ImageWindow struct.
 */
static void
cb_toolbar_open_button (GtkWidget *widget)
{
   create_filebrowser (NULL);
}


/*
 *  cb_toolbar_prefs_button:
 *     @ Callback function for prefs button.
 *     @ Open preference button. 
 *
 *  widget :
 */
static void
cb_toolbar_prefs_button (GtkWidget *widget)
{
   prefs_open_window();
}


static void
cb_toolbar_prev_button  (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   imageview_prev (iw->iv);
}


static void
cb_toolbar_next_button  (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   imageview_next (iw->iv);
}


/*
 *  cb_toolbar_no_zoom:
 *     @ Callback function for toolbar no zoom button.
 *       Zoom image.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_toolbar_no_zoom (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   imageview_zoom (iw->iv, ZOOM_100, 0, 0);
}


/*
 *  cb_toolbar_zoom_in:
 *     @ Callback function for toolbar zoom in button.
 *       Zoom image.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_toolbar_zoom_in (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   imageview_zoom (iw->iv, ZOOM_IN, 0, 0);
}


/*
 *  cb_toolbar_zoom_out:
 *     @ Callback function for toolbar zoom out button.
 *       Zoom image.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_toolbar_zoom_out (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   imageview_zoom (iw->iv, ZOOM_OUT, 0, 0);
}


/*
 *  cb_toolbar_zoom_fit:
 *     @ Callback function for toolbar zoom fit button.
 *       Zoom image.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void cb_toolbar_zoom_fit (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   imageview_zoom (iw->iv, ZOOM_FIT, 0, 0);
}


/*
 *  cb_toolbar_zoom:
 *     @ Callback function for toolbar zoom button.
 *       Zoom image.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_toolbar_zoom (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   iw->iv->x_scale = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(iw->button.xscale));
   iw->iv->y_scale = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(iw->button.yscale));
   imageview_show_image (iw->iv, iw->iv->rotate);
}


/*
 *  cb_toolbar_rotate:
 *     @ Callback function for toolbar rotate button.
 *       Rotate image.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_toolbar_rotate (GtkWidget *widget, ImageWindow *iw)
{
   guint angle;

   g_return_if_fail (iw);

   /* convert to absolute angle */
   angle = ( iw->iv->rotate + 3 ) % 4;

   imageview_show_image (iw->iv, angle);
}


/*
 *  cb_toolbar_keep_aspect:
 *     @ Callback function for toolbar spin button.
 *     @ Set y_scale value to x_scale value if keep aspect.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_toolbar_keep_aspect (GtkWidget *widget, ImageWindow *iw)
{
   gfloat scale;

   g_return_if_fail (iw);

   if (!iw->iv->keep_aspect)
      return;

   scale = gtk_spin_button_get_value_as_float (GTK_SPIN_BUTTON (iw->button.xscale));
   gtk_spin_button_set_value (GTK_SPIN_BUTTON (iw->button.yscale), scale);
}


/*
 *  cb_toolbar_fit_window:
 *     @ Callback function for toolbar fit window button.
 *     @ Fit window size to image.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_toolbar_fit_window (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   imagewin_set_win_size (iw);
}


/*
 *  cb_toolbar_fullscreen:
 *     @ Callback function for toolbar fullscreen button.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_toolbar_fullscreen (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);
   menu_check_item_set_active (iw->view_menu, "/Full Screen", !iw->fullscreen);
}


static gint
cb_scale_spinner_key_press (GtkWidget *widget, 
			    GdkEventKey *event,
			    ImageWindow *iw)
{
   g_return_val_if_fail (iw, FALSE);
        
   switch (event->keyval) {
   case GDK_Escape:
      gtk_window_set_focus (GTK_WINDOW (iw->window), NULL);
      return TRUE;
   }

   return TRUE;
}


/******************************************************************************
 *
 *  other callback functions.
 *
 ******************************************************************************/
/*
 *  cb_window_closed:
 *     @ Callback function for Image Window destroy event.
 *
 *  widget :
 *  im     : Pointer to ImageWindow structure.
 */
static void
cb_window_closed (GtkWidget *widget, ImageWindow *iw)
{
   g_return_if_fail (iw);

   if (g_list_length (ImageWinList) == 1 && conf.imgwin_save_win_state)
      imagewin_store_win_state_to_config (iw);

   gtk_signal_disconnect_by_func (GTK_OBJECT (iw->iv),
				  (GtkSignalFunc) cb_set_list,   iw);
   gtk_signal_disconnect_by_func (GTK_OBJECT (iw->iv),
				  (GtkSignalFunc) cb_unset_list, iw);

   iw->iv = NULL;

   /* update linked list */
   ImageWinList = g_list_remove (ImageWinList, iw);

   if (iw == shared_img_win)
      shared_img_win = NULL;

   /* free mem for image view window info */
   g_free(iw);

   /* quit when last window */
   if (!g_list_first(ImageWinList) && !g_list_first(ThumbWinList)) {
      gimageview_quit();
   }
}


static void
cb_set_list (ImageView *iv, ImageWindow *iw)
{
   g_return_if_fail (iw);
   g_return_if_fail (IS_IMAGEVIEW (iv));

   gtk_widget_set_sensitive (iw->button.prev, TRUE);
   gtk_widget_set_sensitive (iw->button.next, TRUE);

   menu_item_set_sensitive (iw->move_menu, "/Next",     TRUE);
   menu_item_set_sensitive (iw->move_menu, "/Previous", TRUE);
   menu_item_set_sensitive (iw->move_menu, "/First",    TRUE);
   menu_item_set_sensitive (iw->move_menu, "/Last",     TRUE);
}


static void
cb_unset_list (ImageView *iv, ImageWindow *iw)
{
   g_return_if_fail (iw);
   g_return_if_fail (IS_IMAGEVIEW (iv));

   gtk_widget_set_sensitive (iw->button.prev, FALSE);
   gtk_widget_set_sensitive (iw->button.next, FALSE);

   menu_item_set_sensitive (iw->move_menu, "/Next",     FALSE);
   menu_item_set_sensitive (iw->move_menu, "/Previous", FALSE);
   menu_item_set_sensitive (iw->move_menu, "/First",    FALSE);
   menu_item_set_sensitive (iw->move_menu, "/Last",     FALSE);
}


static void
cb_rendered (ImageView *iv, ImageWindow *iw)
{
   g_return_if_fail (iw);
   if (!g_list_find (ImageWinList, iw)) return;

   g_return_if_fail (iw->iv);
   if (!g_list_find (ImageViewList, iw->iv)) return;

   /* set statu bar */
   imagewin_set_statusbar_info (iw);

   /* set title */
   imagewin_set_window_title (iw);

   /* resize window */
   if (!iw->fullscreen && conf.imgwin_fit_to_image)
      imagewin_set_win_size (iw);

   if (iw->button.xscale)
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (iw->button.xscale),
				 iw->iv->x_scale);
   if (iw->button.yscale)
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (iw->button.yscale),
				 iw->iv->y_scale);
}


static void
cb_toggle_aspect (ImageView *iv, gboolean keep_aspect, ImageWindow *iw)
{
   g_return_if_fail (iw);

   gtk_widget_set_sensitive (iw->button.yscale, !keep_aspect);
}



/******************************************************************************
 *
 *   public functions.
 *
 ******************************************************************************/
void
imagewin_player_set_sensitive_all (ImageWindow *iw, gboolean sensitive)
{
   g_return_if_fail (iw);

   gtk_widget_set_sensitive (iw->player.prev,    sensitive);
   gtk_widget_set_sensitive (iw->player.rw,      sensitive);
   gtk_widget_set_sensitive (iw->player.play,    sensitive);
   gtk_widget_set_sensitive (iw->player.pause,   sensitive);
   gtk_widget_set_sensitive (iw->player.stop,    sensitive);
   gtk_widget_set_sensitive (iw->player.fw,      sensitive);
   gtk_widget_set_sensitive (iw->player.next,    sensitive);
   gtk_widget_set_sensitive (iw->player.eject,   sensitive);
   gtk_widget_set_sensitive (iw->player.seekbar, sensitive);
}


void
imagewin_change_image_file (ImageWindow *iw, const gchar *newfile)
{
   GList *node;

   g_return_if_fail (iw);

   node = g_list_find (ImageWinList, iw);
   if (!node) return;

   imageview_change_image_file (iw->iv, newfile);

   if (conf.imgwin_raise_window)
      gdk_window_raise (iw->window->window);

   if (iw->slideshow)
      iw->iv->fit_to_frame = conf.slideshow_fit_to_win;

   if (iw->iv->fit_to_frame && conf.imgwin_fit_to_image && !iw->fullscreen)
      gdk_window_resize (iw->window->window, conf.imgwin_width, conf.imgwin_height);

   imageview_show_image(iw->iv, iw->iv->rotate);
}


void
imagewin_store_win_state_to_config (ImageWindow *iw)
{
   g_return_if_fail (iw && iw->window);

   if (iw->slideshow) return;

   if (!iw->fullscreen) {
      conf.imgwin_width  = iw->window->allocation.width;
      conf.imgwin_height = iw->window->allocation.height;
   } else {
      conf.imgwin_width  = iw->win_width;
      conf.imgwin_height = iw->win_height;
   }

   conf.imgwin_show_menubar   = iw->show_menubar;
   conf.imgwin_show_toolbar   = iw->show_toolbar;
   conf.imgwin_show_player    = iw->show_player;
   conf.imgwin_show_statusbar = iw->show_statusbar;
   conf.imgview_scrollbar     = iw->iv->show_scrollbar;
}


/*
 *   imagewin_initialize:
 *      @ allocate new ImageWindow & ImageView struct, and initialize it.
 *
 *   filename : image file name for open.
 *   Return   : Pointer to ImageWindow struct.
 */
ImageWindow *
imagewin_initialize (const gchar *filename)
{
   ImageWindow *iw;
   ImageView   *iv;

   iv = IMAGEVIEW (imageview_new (filename));
   if (!iv)
      return NULL;

   iw = g_new0 (ImageWindow, 1);
   if (!iw)
      return NULL;

   iw->iv = iv;
   iw->window_creating    = TRUE;
   iw->show_menubar       = conf.imgwin_show_menubar;
   iw->show_toolbar       = conf.imgwin_show_toolbar;
   iw->show_player        = conf.imgwin_show_player;
   iw->show_statusbar     = conf.imgwin_show_statusbar;
   iw->fullscreen         = FALSE;
   iw->hide_frame         = FALSE;

   ImageWinList = g_list_append (ImageWinList, iw);

   return iw;
}


/*
 *   imagewin_create_widnow.
 *      @ Create new iamge window. 
 *        ImageWindow struct must be initialized before call this function.
 *
 *   iw     : Pointer to ImageWindow structure.
 *   Return : Return TRUE if success.
 */
gboolean
imagewin_create_window (ImageWindow *iw)
{
   GtkWidget *vbox, *hbox;
   GtkObject *adj;
   Icon *icon;

   if (!iw)
      return FALSE;

   /* Image Window */
   iw->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   gtk_window_set_wmclass(GTK_WINDOW(iw->window), "imagewin", "GImageView");
   gtk_widget_realize (iw->window);
   if (iw->hide_frame)
      gdk_window_set_decorations (iw->window->window, 0);
   gtk_window_set_policy(GTK_WINDOW(iw->window), TRUE, TRUE, FALSE);
   gtk_window_set_default_size (GTK_WINDOW(iw->window),
				conf.imgwin_width, conf.imgwin_height);
   gtk_signal_connect (GTK_OBJECT (iw->window), "destroy",
		       GTK_SIGNAL_FUNC (cb_window_closed), iw);
   imagewin_set_window_title (iw);

   gtk_widget_show (iw->window);

   icon = icon_get ("gimv_icon");
   if (icon)
      gdk_window_set_icon(iw->window->window, NULL, icon->pixmap, icon->mask);

   /* Main vbox */
   vbox = gtk_vbox_new (FALSE, 0);
   iw->main_vbox = vbox;
   if (iw == shared_img_win)
      gtk_widget_set_name (iw->main_vbox, "SharedImageWin");
   else
      gtk_widget_set_name (iw->main_vbox, "ImageWin");
   gtk_container_add (GTK_CONTAINER (iw->window), vbox);
   gtk_widget_show (vbox);

   /* Menu Bar */
   iw->menubar_handle = gtk_handle_box_new();
   gtk_widget_set_name (iw->menubar_handle, "MenuBarContainer");
   gtk_container_set_border_width (GTK_CONTAINER(iw->menubar_handle), 2);
   gtk_box_pack_start(GTK_BOX(vbox), iw->menubar_handle, FALSE, FALSE, 0);

   if (!iw->show_menubar)
      gtk_widget_hide (iw->menubar_handle);
   else
      gtk_widget_show (iw->menubar_handle);

   /* toolbar */
   iw->toolbar_handle = gtk_handle_box_new();
   gtk_widget_set_name (iw->toolbar_handle, "ToolBarContainer");
   gtk_container_set_border_width (GTK_CONTAINER(iw->toolbar_handle), 2);
   gtk_box_pack_start(GTK_BOX(vbox), iw->toolbar_handle, FALSE, FALSE, 0);
   iw->toolbar = create_toolbar(iw, iw->toolbar_handle);
   gtk_container_add(GTK_CONTAINER(iw->toolbar_handle), iw->toolbar);
   gtk_widget_show_all (iw->toolbar);

   gtk_toolbar_set_style (GTK_TOOLBAR(iw->toolbar), conf.imgwin_toolbar_style);

   if (!iw->show_toolbar)
      gtk_widget_hide (iw->toolbar_handle);
   else
      gtk_widget_show (iw->toolbar_handle);

   /* player toolbar */
   iw->player_handle = gtk_handle_box_new();
   gtk_widget_set_name (iw->player_handle, "PlayerToolBarContainer");
   gtk_container_set_border_width (GTK_CONTAINER(iw->player_handle), 2);
   gtk_box_pack_start(GTK_BOX(vbox), iw->player_handle, FALSE, FALSE, 0);

   hbox = gtk_hbox_new (FALSE, 0);
   gtk_container_add(GTK_CONTAINER(iw->player_handle), hbox);
   gtk_widget_show (hbox);

   iw->player_bar = create_player_toolbar(iw, iw->player_handle);
   gtk_box_pack_start (GTK_BOX (hbox), iw->player_bar, FALSE, FALSE, 0);
   gtk_widget_show_all (iw->player_bar);

   adj = gtk_adjustment_new (0.0, 0.0, 100.0, 0.1, 1.0, 1.0);

   /* draw area */
   gtk_widget_show (GTK_WIDGET (iw->iv));
   gtk_box_pack_start (GTK_BOX (iw->main_vbox), GTK_WIDGET (iw->iv),
		       TRUE, TRUE, 0 );
   gtk_signal_connect (GTK_OBJECT (iw->iv), "set_list",
		       GTK_SIGNAL_FUNC (cb_set_list), iw);
   gtk_signal_connect (GTK_OBJECT (iw->iv), "unset_list",
		       GTK_SIGNAL_FUNC (cb_unset_list), iw);
   gtk_signal_connect (GTK_OBJECT (iw->iv), "rendered",
		       GTK_SIGNAL_FUNC (cb_rendered), iw);
   gtk_signal_connect (GTK_OBJECT (iw->iv), "toggle_aspect",
		       GTK_SIGNAL_FUNC (cb_toggle_aspect), iw);

   iw->player.seekbar = gtk_hscale_new (GTK_ADJUSTMENT (adj));
   gtk_scale_set_draw_value (GTK_SCALE (iw->player.seekbar), FALSE);
   gtk_box_pack_start (GTK_BOX (hbox), iw->player.seekbar, TRUE, TRUE, 0);
   gtk_widget_show (iw->player.seekbar);

   gtk_toolbar_set_style (GTK_TOOLBAR(iw->player_bar), conf.imgwin_toolbar_style);

   if (!iw->show_player)
      gtk_widget_hide (iw->player_handle);
   else
      gtk_widget_show (iw->player_handle);

   /* status bar */
   hbox = gtk_hbox_new (FALSE, 0);
   iw->status_bar_container = hbox;
   gtk_box_pack_end (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   gtk_widget_show (hbox);

   iw->status_bar1 = gtk_statusbar_new ();
   gtk_widget_set_name (iw->status_bar1, "StatuBar1");
   gtk_container_border_width (GTK_CONTAINER (iw->status_bar1), 1);
   gtk_widget_set_usize(iw->status_bar1, 200, -1);
   gtk_box_pack_start (GTK_BOX (hbox), iw->status_bar1, TRUE, TRUE, 0);
   gtk_statusbar_push(GTK_STATUSBAR (iw->status_bar1), 1, "New Window");
   gtk_widget_show (iw->status_bar1);

   iw->status_bar2 = gtk_statusbar_new ();
   gtk_widget_set_name (iw->status_bar1, "StatuBar2");
   gtk_container_border_width (GTK_CONTAINER (iw->status_bar2), 1);
   gtk_widget_set_usize(iw->status_bar2, 50, -1);
   gtk_box_pack_end (GTK_BOX (hbox), iw->status_bar2, TRUE, TRUE, 0);
   gtk_widget_show (iw->status_bar2);

   imagewin_player_set_sensitive_all (iw, FALSE);

   if (!iw->show_statusbar)
      gtk_widget_hide (iw->status_bar_container);
   else
      gtk_widget_show (iw->status_bar_container);

   /* create menus */
   create_imageview_menus (iw);

   while (gtk_events_pending ()) gtk_main_iteration ();

   /* Show Image */
   imageview_show_image(iw->iv, iw->iv->rotate);

   iw->window_creating    = FALSE;

   gtk_widget_grab_focus (iw->iv->draw_area);

   return TRUE;
}


/*
 *   imagewin_open_window.
 *      @ Initialize ImageWindow & ImageView struct and create new image window. 
 *
 *   filename : Image file name for open.
 *              If this value is NULL, create blank canvas.
 *   Return   : New ImageWindow struct.
 */
ImageWindow *
imagewin_open_window (const gchar *filename)
{
   ImageWindow *iw;

   iw = imagewin_initialize (filename);
   imagewin_create_window (iw);

   return iw;
}


ImageWindow *
imagewin_open_shared_window (const gchar *filename)
{
   ImageWindow *iw;

   if (shared_img_win) {
      iw = shared_img_win;
      imagewin_change_image_file (iw, filename);
   } else {
      iw = imagewin_initialize (filename);
      shared_img_win = iw;
      imagewin_create_window (iw);
   }

   return iw;
}


ImageWindow *
imagewin_open_window_auto (const gchar *filename)
{
   ImageWindow *iw;

   if (conf.imgwin_open_new_win) {
      iw = imagewin_open_window (filename);
   } else {
      iw = imagewin_open_shared_window (filename);
   }

   return iw;
}
